/**
 * apeman app to combine multiple handlers into one.
 * @memberof module:apeman-app-combine/lib
 * @function apemanAppCombine
 * @param {function} handlers - Request handlers to combine.
 * @param {object} [options] - Optional settings.
 * @returns {function} - Defined app function.
 */

'use strict'

const async = require('async')
const co = require('co')

const parseFunction = require('parse-function')
const debug = require('debug')('apeman:app:combine')

const GeneratorFunction = Object.getPrototypeOf(function * () {})

let isErrorHandler = (handler) => parseFunction(handler).args[ 0 ] === 'err'
let isGenerator = (fn) => Object.getPrototypeOf(fn) === GeneratorFunction

/** @lends create */
function create (handlers, options = {}) {
  handlers = [].concat(handlers)
  let normalHandlers = handlers.filter((handler) => !isErrorHandler(handler))
  let errorHandlers = handlers.filter((handler) => isErrorHandler(handler))

  /**
   * Defined app.
   * @function app
   * @param {object} req - Request object.
   * @param {object} res - Response object.
   * @param {function} next - Pass to next handler.
   */
  function app (req, res, next) {
    async.eachSeries(normalHandlers, (handler, next) => {
      if (isGenerator(handler)) {
        co(handler, req, res, next).catch(next)
      } else {
        handler(req, res, next)
      }
    }, next)
  }

  // Description of this app.
  Object.assign(app, {
    $desc: 'Combine multiple handlers into one.',
    error (err, req, res, next) {
      async.waterfall([
        (callback) => callback(null, err),
        ...errorHandlers.map((errorHandler) => function (err, callback) {
          let next = (err) => callback(null, err || null)
          if (isGenerator(errorHandler)) {
            co(errorHandler, err, req, res, next).catch(next)
          } else {
            errorHandler(err, req, res, next)
          }
        })
      ], (asyncErr, err) => next(err))
    }
  })

  return [
    app,
    app.error
  ]
}

Object.assign(create, {
  isErrorHandler,
  isGenerator
})

module.exports = create
